"""
This module provides functions for conducting PDF file checks and handling their responses.

The main feature is to extract tables with the following headings.

RuleId  Description Status


"""

from typing import Generator, Sequence

import camelot  # type: ignore
import pandas as pd

from .ten8t_exception import Ten8tException
from .ten8t_result import TR
from .ten8t_util import str_to_bool


def extract_tables_from_pdf(file_path: str,
                            required_columns: Sequence[str] | None = None,
                            pages: str = 'all') -> list[pd.DataFrame]:
    """
    Extracts tables from a PDF file that include specified columns, and returns them in a list.


        | RuleID     | Status | Skip  | None          |
        |------------|--------|-------|---------------|
        | Rule ID 1  | pass   | no    | Rule desc 1   |
        | Rule ID 2  | fail   | yes   | Rule desc 2   |
        |    ...     |  ...   |  ...  |     ...       |

    Args:
        file_path: The path to the PDF file.
        required_columns: The list of columns that must be present in the table.
            Defaults to ["RuleId", "Note", "Status"].
        pages: The pages of the PDF to process. It could be 'all', a range like '1-7', or
            specific pages like '1,3,5'. Defaults to 'all'.

    Returns:
        A list of pandas DataFrames representing the tables extracted from the PDF file.
    """
    required_columns = required_columns or ["RuleId", "Note", "Status"]
    filtered_tables = []
    try:
        tables = camelot.read_pdf(file_path, flavor='stream', pages=pages)
        for table in tables:
            # The table that camelot returns seems to have the columns in
            # the first row, this code moves those values into column names.
            df = table.df
            df.columns = df.iloc[0]
            df = df.drop(df.index[0])
            if set(required_columns).issubset(df.columns):
                filtered_tables.append(df)
    except IOError as ioex:
        raise Ten8tException(f"IO Error reading PDF file {file_path}.") from ioex
    except Exception as exc:
        raise Ten8tException(f"Error extracting tables from PDF: {file_path}") from exc

    return filtered_tables


DEFAULT_COL_NAMES = {
    'status_col': 'Status',
    'note_col': 'Note',
    'rule_col': 'RuleId',
    'skip_col': 'Skip',
}


def rule_from_pdf_rule_ids(file_path: str,
                           rule_id: str,
                           default_msg: str | None = None,
                           col_names: dict | None = None,
                           max_results: int = 1,
                           pages="all") -> Generator[TR, None, None]:
    """
        Yield matching rule ID results from tables contained in a specified PDF file.


        The tables within the PDF file are expected to have the following columns structure with these default names:

        Reads tables from a specified PDF file and inspects each table for data that matches
        the supplied `rule_id`. For each match found, the function extracts the corresponding status, note,
        and skip values, and yields these in a namedtuple. If multiple tables exist or the `rule_id` does not
        match any table, the function continues execution.

        Parameters:
        ----------
        file_path : str
           Path to the PDF file to extract tables from.
        rule_id : str
           Specific rule ID to search for within the tables.
        default_msg : str, optional
           Default message to return with the rule result. If provided, it overrides column name information.
           Defaults to None.
        col_names : dict, optional
           Dictionary for renaming column names. Defaults to None.
        max_results : int, optional
           Maximum number of results to yield. The function raises an exception when exceed this number. Defaults to 1.
        pages : str
           Specific pages from the PDF file to be read. Defaults to 'all'.

        Yields:
        -------
        namedtuple
           Namedtuple represented by BR where:
           `status` (bool) : Indicates 'pass' if True and 'fail' if False.
           `msg` (str) : Contains messages or notes associated with the rule.
           `skipped` (bool) : Indicates if the rule was skipped. True if yes, False otherwise.

        Raises:
        -------
        Ten8tException
           If the number of yielded results exceeds the maximum limit specified by `max_results`.
        """
    col_names = DEFAULT_COL_NAMES | col_names if col_names else DEFAULT_COL_NAMES
    status_col = col_names['status_col']
    note_col = col_names['note_col']
    rule_col = col_names['rule_col']
    skip_col = col_names['skip_col']

    tables = extract_tables_from_pdf(file_path,
                                     required_columns=[rule_col, note_col, status_col],
                                     pages=pages)
    count = 0
    for table in tables:
        if rule_col in table.columns:
            df = table[table[rule_col] == rule_id]
            for _, rule_data in df.iterrows():
                status = str_to_bool(rule_data[status_col])
                skip = rule_data.get(skip_col, False)
                msg = default_msg or rule_data[note_col]
                yield TR(status=status, msg=msg, skipped=skip)
                count += 1
                if count > max_results:
                    raise Ten8tException(f"Maximum number of results ({max_results}) exceeded for {file_path}")
    if count == 0:
        yield TR(status=None, skipped=True, msg=f"No results found in {file_path}")



"""
This module provides functions for using ping to verify network connectivity.


"""
from typing import Generator

import ping3

from .ten8t_format import BM
from .ten8t_result import TR

NO_HOSTS_MSG = "No hosts provided for ping."
MIN_LATENCY_MS = 0.0001


def rule_ping_check(hosts: str | list, timeout_ms: float = 4000.0, skip_on_none=False, pass_on_none=False) -> Generator[
    TR, None, None]:
    """
    Given a sequence of hosts perform ping checks against each of them given a single timeout
    value.  This function handles the following:
    hosts = "google.com"   converted to ["google.com"]
    hosts = "google.com apple.com"   converted to ["google.com","apple.com"]
    hosts = ["google.com"]   as is
    Args:
        hosts: string with list of hosts or list of hosts
        timeout_ms(4000): Time in milliseconds.
        skip_on_none(False): If true an empty host list is a skip
        pass_on_none(False): If try the tests passes on empty host list
    Yields:
        list of results
    """
    hosts = hosts.replace(',', ' ').split() if isinstance(hosts, str) else hosts

    if len(hosts) == 0:
        if skip_on_none:
            yield TR(status=True, skipped=True, msg=NO_HOSTS_MSG)
            return
        else:
            yield TR(status=pass_on_none, msg=NO_HOSTS_MSG)
            return

    for host in hosts:
        try:
            # All of this to make this call
            latency = ping3.ping(host, timeout=timeout_ms, unit='ms')
            timeout_str = f'{timeout_ms:0.1f}'
            if latency is False:
                yield TR(status=False,
                         msg=f"No ping response from server {BM.code(host)} timeout = {BM.code(timeout_str)} ms")
            elif latency < MIN_LATENCY_MS:
                latency_str = f"{MIN_LATENCY_MS:0.1f}"
                yield TR(status=True, msg=f"Host {BM.code(host)} is up: response time < {BM.code(latency_str)} ms")
            else:
                latency_str = f"{latency:0.1f}"
                yield TR(status=True, msg=f"Host {BM.code(host)} is up, response time = {BM.code(latency_str)} ms")
        except Exception:
            # Document your exception handling logic here
            yield TR(status=False, msg=f"Host {BM.code(host)} not found: {BM.fail(host)}")



"""
This module provides functions for conducting URL checks and handling their responses.

Includes:
- `rule_url_200`: Checks if an array of URLs returns a specific HTTP status code.

- `is_mismatch`: Compares two dictionaries, returns the first differing pair.

- `rule_web_api`: Verifies a URLs HTTP response status code and compares returned JSON data.

Uses the `requests` library for HTTP requests, and handles exceptions accordingly.
"""
from typing import Generator, Sequence

import requests
from requests.exceptions import RequestException

from . import Ten8tException
from .ten8t_format import BM
from .ten8t_result import TR, Ten8tYield


def rule_url_200(urls: str | Sequence[str],
                 expected_status=200,
                 timeout_sec=5,
                 summary_only=False,
                 summary_name=None) -> Generator[TR, None, None]:
    """
    Simple rule check to verify that URL is active.
    
    TAs setup this just checks that a web API URL is active. The urls parameter
    is set up to be tolerant of varius ways you could pass a list of URLS
    
    urls(str|Sequence[str]): List of URLs to check or a string that can be split into URLs
    expected_status(200): Expected response status code
    timeout_sec(int): Timeout in seconds.
    
    """
    y = Ten8tYield(summary_only=summary_only, summary_name=summary_name or "Rule 200 Check")

    # Allow strings to be passed in
    if isinstance(urls, str):
        urls = urls.replace(",", " ").split()

    for url in urls:
        try:
            response = requests.get(url, timeout=timeout_sec)
            url_str = BM.code(url)
            code_str = BM.code(response.status_code)

            if response.status_code == expected_status:
                yield from y(status=True, msg=f"URL {url_str} returned {code_str}")
            else:
                yield from y(
                    status=response.status_code == expected_status,
                    msg=f"URL {url_str} returned {code_str}",
                )

        except RequestException as ex:
            yield from y(status=False, msg=f"URL{BM.code(url)} exception.", except_=ex)

    if summary_only:
        yield from y.yield_summary()


def is_mismatch(dict1, dict2):
    """
    Return the first differing values from dict1 and dict2
    Args:
        dict1:
        dict2:

    Returns: None if every key/value pair in dict1 is in dict, otherwise
            returns the first value that differs from dict 2

    """
    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        return False
    for key, value in dict1.items():
        if key not in dict2:
            return {key: value}
        if isinstance(value, dict):
            nested_result = is_mismatch(value, dict2[key])
            if nested_result is not None:  # Manual short-circuit the mismatch search.
                return {key: nested_result}
        elif value != dict2[key]:
            return {key: value}
    return None  # Return None if it is a subset.


def rule_web_api(url: str,
                 json_d: dict,
                 timeout_sec=5,
                 expected_response=200,
                 timeout_expected=False,
                 summary_only=False,
                 summary_name=None) -> Generator[TR, None, None]:
    """Simple rule check to verify that URL is active and handles timeouts."""
    y = Ten8tYield(summary_only=summary_only)
    try:

        if isinstance(expected_response, int):
            expected_response = [expected_response]
        elif isinstance(expected_response, str):
            expected_response = str.split(expected_response)
        elif not isinstance(expected_response, list):
            raise Ten8tException(f"Expected integer or list of integers for 'expected_response' for {url}")

        response = requests.get(url, timeout=timeout_sec)

        if response.status_code not in expected_response:
            yield from y(status=False,
                         msg=f"URL {BM.code(url)} expected {BM.expected(expected_response)} returned {BM.actual(response.status_code)} ")
            return

        # This handles an expected failure by return true but not checking the json
        if response.status_code != 200:
            yield from y(status=True,
                         msg=f"URL {BM.code(url)} returned {BM.code(response.status_code)}, no JSON comparison needed.")
            return

        response_json: dict = response.json()
        # d_status = verify_dicts(response_json, json_d)

        d_status = is_mismatch(json_d, response_json)

        if d_status is None:
            yield from y(status=True,
                         msg=f"URL {BM.code(url)} returned the expected JSON {BM.code(json_d)}")
        else:
            yield from y(status=False,
                         msg=f"URL {BM.code(url)} did not match at key {d_status}")

    except (requests.exceptions.ReadTimeout, requests.exceptions.Timeout):  # pragma: no cover
        yield from y(status=timeout_expected, msg=f"URL {BM.code(url)} timed out.")

    if summary_only:
        yield from y.yield_summary(summary_name or "Rule 200 check")
